/* NETLOG IP/tcp_port logger. Log data from a tcp source to file.
 * sivann - Spiros Ioannou
 * v0.1 2/2010 first version
 *
 */

/*uncomment the following to log each connection to syslog*/
/* #define  LOGCONN   */


#include <sys/types.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <sys/socket.h>

#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <poll.h>


#include <stropts.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <signal.h>
#include <stdio.h>
#include <strings.h>
#include <stdlib.h>
#include <syslog.h>


char            buffer[100000];
long int        lsd, b,r, i, Asd;
socklen_t 	len;
int             debug=1;
struct sockaddr_in sa;
unsigned char   in[2];
struct timeval  tval;
fd_set          inout_fds;
char 		errstr[256];

int 		mypid=0;

int priority_e = LOG_ERR | LOG_USER;
int priority_i = LOG_INFO | LOG_USER;
int priority_w = LOG_WARNING | LOG_USER;
int priority_d = LOG_DEBUG | LOG_USER;

int my_usleep(long x);

void daemonize()
{
  int rc;
  int fd;

  fflush(stdout);
  fflush(stderr);

  rc = fork();
  if(rc < 0) {
      sprintf (errstr,"fork:%s",strerror(errno));
      printf("%s",errstr);
      syslog(priority_e,errstr);
      exit(1);
  }

  if(rc > 0)
      exit(0);

  close(0);
  close(1);
  close(2);
  /* Leaving the default file descriptors free is not a good
     idea, as it will cause library functions such as abort to
     thrash the on-disk cache. */
  fd = open("/dev/null", O_RDONLY);
  if(fd > 0) {
      dup2(fd, 0);
      close(fd);
  }
  fd = open("/dev/null", O_WRONLY);
  if(fd >= 0) {
      if(fd != 1)
	  dup2(fd, 1);
      if(fd != 2)
	  dup2(fd, 2);
      if(fd != 1 && fd != 2)
	  close(fd);
  }

  rc = setsid();
  if(rc < 0) {
      sprintf (errstr,"setsid:%s",strerror(errno));
      printf("%s",errstr);
      syslog(priority_e,errstr);
      exit(1);
  }
}



int main(int argc, char *argv[])
{
  int             one = 1;
  int 	fd;
  int logopt = LOG_PID | LOG_CONS;
  int facility = LOG_USER;
  char	ident[256];
  char  fname[1024];
  int port;

  if (argc < 3) {
    printf("\nnetlog, Spiros Ioannou, v0.1\n");
    printf("Usage:\t netlog <localport> <localfile> \n");
    printf("or:\t netlog <localport> <localfile> debug\n\n");
    printf("Logs data from tcp port <localport> to <localfile>.\n");
    printf("With the 'debug', verbose output is generated\n\n");
    printf("Without the 'debug', process runs as daemon\n\n");
    exit(1);
  }

  openlog(ident,logopt, facility);

  if ((argc==4) && strstr(argv[3],"debug")) {
    debug=1;
    setvbuf(stdout, NULL, _IONBF, 0); //turn off buffering for stdout
  }
  else {
    debug=0;
    daemonize();
  }



  mypid=getpid();

  sprintf(ident,"%s[%d]",argv[0],mypid);

  if (debug)
    printf("Debug ON\n");

  port=atoi(argv[1]);
  if (port<=0) {
    printf("Invalid port %d (%s)",port,argv[1]);
    exit(2);
  }

  strcpy(fname,argv[2]);

  sprintf (errstr,"netlog starting port:%d, file:%s\n",port,fname);
  syslog(priority_i,errstr);

  fd=open(fname,O_APPEND|O_CREAT|O_WRONLY,0644);
  if (fd<0) {
    sprintf (errstr,"%s:open:%s\n",fname,strerror(errno));
    printf("%s",errstr);
    syslog(priority_e,errstr);
    exit(3);
  }

  /* Open  the "listening"  socket */
  if ((lsd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
    sprintf (errstr,"lsd:socket:%s\n",strerror(errno));
    printf("%s",errstr);
    syslog(priority_e,errstr);
    exit(1);
  }
  setsockopt(lsd, SOL_SOCKET, SO_REUSEADDR, (char *) &one, sizeof(one));
  sa.sin_family = AF_INET;
  sa.sin_port = htons((u_short) port);
  sa.sin_addr.s_addr = htonl(INADDR_ANY);

  if (bind(lsd, (struct sockaddr *) & sa, sizeof(sa)) < 0) {
    sprintf (errstr,"lsd:port [%d]: bind:%s\n",port,strerror(errno));
    printf("%s",errstr);
    syslog(priority_e,errstr);
    exit(2);
  }

  if (listen(lsd, 2) < 0) { //backlog=2. This is not a multi-client server.
    sprintf (errstr,"lsd:listen:%s\n",strerror(errno));
    printf("%s",errstr);
    syslog(priority_e,errstr);
    exit(errno);
  }
  len = sizeof(sa);
  strcpy(buffer, "");

  if (debug)
    printf("daemon ready ...\n");

  Asd=0;

  while (1) {
    fflush(stdout);
    tval.tv_sec = 0;
    tval.tv_usec = 0;		/* 50 microseconds */


    FD_ZERO(&inout_fds);	/* Initialize file descriptors for select in loop */

    /* Prepare for input from connected host A */
    if (Asd > 0)
      FD_SET(Asd, &inout_fds);

    /* Else prepare to accept a connection to our socket */
    if (!Asd)
      FD_SET(lsd, &inout_fds);


    if (debug)
	printf("%d:before select\n",mypid);

    /* Do a non-timeout select (interupt driven) to minimize load */
    do {
      i = select(64, &inout_fds, (fd_set *) 0, (fd_set *) 0,
		 (struct timeval *) 0 /* &tval */ );
      if (i < 0) {
	sprintf (errstr,"select %s\n",strerror(errno));
	syslog(priority_w,errstr);
	if (debug)
	  printf("%s",errstr);
      }

    } while ((i == EINTR) || (i < 0));

    if (debug)
      printf("%d:after select(=%ld)\n",mypid, i);


    /* A new connection, ... */
    if ((!Asd) && FD_ISSET(lsd, &inout_fds) ) {
      if ((Asd = accept(lsd, (struct sockaddr *) & sa, &len)) < 0) {
	sprintf (errstr,"accept Asd %s\n",strerror(errno));
	syslog(priority_w,errstr);
	FD_ZERO(&inout_fds);	
	Asd=0;
	my_usleep(300000); //300ms
	if (debug)
	  printf("%s",errstr);
	//exit(6);
      }

#ifdef LOGCONN
      sprintf(errstr,"%d:new connection from address:%s,port=%d \n",mypid,
	     inet_ntoa(sa.sin_addr), ntohs(sa.sin_port));
      syslog(priority_d,errstr);
#endif

      if (debug)
	printf("%d:A new connection from address: %s port=%d \n",mypid,
	     inet_ntoa(sa.sin_addr), ntohs(sa.sin_port));
    }

    /* transfer data from A to C */
    if ((Asd > 0) && FD_ISSET(Asd, &inout_fds)) {
      bzero(buffer, sizeof(buffer));

      /* read data from A */
      b = read(Asd, buffer, sizeof(buffer));
      buffer[b]=0;

      if (debug)
	printf("read from socket Asd:#%s#,%ld bytes\n",
	     buffer, b);

      /* If read failed : other end probably closed connection */
      if (b <= 0) {
	/* Free socket Asd for a new connection */
	//perhaps log it only if <0
	shutdown(Asd, 2);
	close(Asd);
	Asd=0;

#ifdef LOGCONN
	sprintf (errstr,"read returned (%d) from client socket (connection closed?): %s\n",b,strerror(errno));
	syslog(priority_d,errstr);
#endif

	if (debug) {
	  printf("%s",errstr);
	  printf("%d:probably closed connection to  address: %s port=%d \n",mypid,
	       inet_ntoa(sa.sin_addr), ntohs(sa.sin_port));
	}
      }
      else { //we read something
	if (debug)
	  printf("\nbuffer written to C is:#%s#=%ld bytes\n",
	       buffer, b);
	
	/* write A data to C */
	r=write(fd, buffer, b);
	if (r<=0) {
	  sprintf (errstr,"write to %s: %s\n",fname,strerror(errno));
	  syslog(priority_e,errstr);
	  if (debug)
	    printf("%s",errstr);
	  exit(errno);
	}
      }//we read something
    } // if Asd>0 && FD_ISSET (Asd)

  }//while (1)
  return 0;
}				/* End of main */


int my_usleep(long x)
{
  struct timeval  time;
  int             r;

  time.tv_sec = x / 1000000;
  time.tv_usec = x % 1000000;
  r = select(0, NULL, NULL, NULL, &time);
  return r;
}
